;DESCRIPTION:
;	Script for AutoHotKey to codificate selected "extensions" files of normal human languaje in a file:
;  			With DOS end of line (CR+LF)  (Optional, not log)
;			Without BOM
;			In char codification Windows-1252 (Windows ANSI).
;CHANGELOG:
;		V.1.0.-2017.11.22: First version
;
;NOTES:
;
;------------------------------------------------------------
;Compilation orders
	#NoEnv  ; Recommended for performance and compatibility with future AutoHotkey releases.
	#Warn  ; Enable warnings to assist with detecting common errors.
	#SingleInstance Off  ;The word OFF allows multiple instances of the script to run concurrently
	;#NoTrayIcon ;Disables the showing of a tray icon.
	SendMode Input  ; Recommended for new scripts due to its superior speed and reliability.
	SplitPath, A_ScriptName,,,, ScriptName ;Get ScriptName for windows title and log file.
	SetWorkingDir %A_ScriptDir%  ; Ensures a consistent starting directory.

;Program behavior and Global variables
	global s_ScriptWinTitle:= StrReplace(ScriptName, "_" , " ", , -1) . " V.1.0"
	global s_LogFile:= A_ScriptDir . "\" . ScriptName . ".log"
	global s_ScriptAction:= "Analyze" 	; Script action: =Analyze (not files changes), =Process overwritting files
	global b_Recursive:= True			; Search recursive in the folders
	global b_DOS_EOLAction:= False		; DOS_end_of_line assuration: Search and remplace (False=Not process, True=Process)
	global b_NoBOMAction:= False    	; No_BOM action: Search and eliminate BOM UTF-8 (True=Not process, False=Process)
	global s_FinalCode:="Nothing"		; Nothing|Win1552|CP850|NoBOM UTF-8|BOM UTF-8
	global b_ForceCodeUnknow:= False   	; After Smart detection force codification with "_autodetect_all" ADO detection.
	global b_ShowDecodeText:= False   	; Show the decoded text for confirm the encode.
	
	global b_DiacriticAndMath:=True		; In the Smart detection: Preserve SP,FR,GE Diacritics and math simbols
	global b_SP_Diacritic:=False		; In the Smart detection: Preserve only Spanish Diacritics.
	
	global b_ChooseCP850:=True			; The Smart detection: Choose CP850 in equal with CP437
	global b_ChooseCP437:=False			; The Smart detection: Choose CP437 in equal with CP850
	
	global b_NewLogFile:= True			; Delete the present Log File and start one new: False= Append, True= Star new one
	global b_VerboseLog:= False			; Log (show) all messages with script action over files: False= No, True= Yes
	global s_TextLog:= ""				; Program Log
	
	global s_WorkPath:=A_ScriptDir		; Working folder for recursive search
	global s_WorkFileExt:="txt,text"	; File extensions implicated.

	global b_Debug:= False				; Full Debug mode. Normaly used for extended verbose messages of large loops for calculate (without action) over file.

;Make the GUI for get the script input parameters and action variables
	Gui 1: -Disabled +SysMenu -Owner  ; -AlwaysOnTop ; +Owner No tray button but provides a normal title bar. +ToolWindow No tray button but provides a narrower title bar.
	Gui 1:Add, Button, x10 y8 w75 Default gGetFolder, Get Folder
	Gui 1:Font, Bold
	Gui 1:Color, cWhite ;Black,Silver,Gray,White,Maroon,Red,Purple,Fuchsia,Green,Lime,Olive,Yellow,Navy,Blue,Teal,Aqua
	Gui 1:Add, Edit, x90 y10 w910 cBlue Border vs_WorkPath, %s_WorkPath%
	Gui 1:Font, Norm

	Gui 1:Add, Button, x10 w75 Default gGetExtensions, Get Extens
	Gui 1:Font, Bold	
	Gui 1:Add, Edit, x90 y38 w910 cBlue Border vs_WorkFileExt, %s_WorkFileExt%
	Gui 1:Font, Norm
	Gui 1:Color, Default

	Gui 1:Add, CheckBox, x10 y68 vb_DOS_EOLAction, DOS End Of Line assurance (CR+LF)?
	Gui 1:Add, CheckBox, x240 y68 vb_NoBOMAction, Not BOM UTF-8 action (search and delete)?
	
	Gui 1:Add, Text, x06 y85 w650 h42 Border, 
	Gui 1:Add, Text, x10 y90, Smart detection for codify to:
	Gui 1:Add, DropDownList, x150 y88 w105 vs_FinalCode Choose1, Nothing|UTF-8_No-BOM|Win1552|CP850|UTF-8_BOM|UTF-16BE_BOM|UTF-16BE_No-BOM|UTF-16LE_BOM|UTF-16LE_No-BOM
	Gui 1:Add, CheckBox, x10 y110 vb_ForceCodeUnknow, Force Unknow file codif.
	Gui 1:Add, CheckBox, x150 y110 vb_ShowDecodeText, Show decode text
	Gui 1: Add, Radio, x265 y90  Group vb_DiacriticAndMath Checked, Preserve SP,FR,GE Diacritics and '  ' simbols
	Gui 1: Add, Radio, x265 y110 vb_SP_Diacritic, Preserve only Spanish Diacritics
	Gui 1: Add, Radio, x515 y90  Group vb_ChooseCP850 Checked, In equals choose CP850
	Gui 1: Add, Radio, x515 y110 vb_ChooseCP437, In equals choose CP437
	
	Gui 1:Add, DropDownList, x670 y68 w75 vs_ScriptAction Choose1, Analyze|Process
	Gui 1:Add, Text, x750 y70, Script action? (Attention, Process overwrite files)
    Gui 1:Add, CheckBox, x735 y90 vb_Recursive Checked, Recursive search in folders?
	
	Gui 1:Add, Text, x10 y130 cBlue, Log Window:	
	Gui 1:Add, Progress, x80 y130 w920 h16 c0FC0FC +Smooth Border vi_ProgBarGlob 
	Gui 1:Add, Text, xp	yp+2 wp	BackgroundTrans +center vs_ProgBarGlobTxt 
	;GuiControl,, i_ProgBarGlob, 100
	GuiControl, Hide, i_ProgBarGlob
	Gui 1:Add, Edit, x10 y148 w990 r17 vs_TextLog, ;-> Edit3 (Search the control name with AutoHotKey Window Spy)
	
	Gui 1:Add, CheckBox, x10 y380 vb_NewLogFile Checked, Delete the present Log File and start one new?
	Gui 1:Add, CheckBox, x800 y380 vb_VerboseLog , Full information log (verbose mode)?
	Gui 1:Add, Button, x10 y400 w75 gScriptProcess, Go!
	Gui 1:Add, Button, x95 y400 w75 gScriptReload, Reload!
	
	Gui 1:Add, Progress, x270 y380 w500 h16 cLime +Smooth Border vi_ProgBarLocal
	Gui 1:Add, Text, xp	yp+2 wp	BackgroundTrans +center vs_ProgBarLocalTxt
	;GuiControl,, i_ProgBarLocal, 100
    GuiControl, Hide, i_ProgBarLocal	
	
	Gui 1:Show , , %s_ScriptWinTitle%
	Return  ;The script wait in the Gui window showed. It continue with the Gui events.

GetFolder:
	;Get folder for recursive search
	FileSelectFolder, s_NewWorkPath, , 2, May you select the folder for recursive search of files?
	if (s_NewWorkPath = "")
		Return
	s_WorkPath:= RegExReplace(s_NewWorkPath, "\\$")  ; Removes the trailing backslash, if present.
	GuiControl,, s_WorkPath, %s_WorkPath% 
Return
GetExtensions:
	;Get File extensions implicated
	Gui 1:Submit, NoHide  ; Save the input from the user to each control's associated variable.
	InputBox,s_WorkFileExt, ,May you introduce the implicated file extensions separated by commas?, ,480,125, , , , ,%s_WorkFileExt%
	GuiControl,, s_WorkFileExt, %s_WorkFileExt%
Return
GuiClose:
GuiEscape:
	Gui 1:Destroy
	;MsgBox, 64, % ScriptName, % ScriptName "-(" . s_ScriptAction . "|" . b_DOS_EOLAction . "|" . b_NoBOMAction . "|" . s_FinalCode . ") canceled.`nGood bye, " . A_UserName . "!" 
	ExitApp
ScriptReload:
	Reload
ScriptProcess:
	Gui 1:Submit, NoHide  ; Save the input from the user to each control's associated variable.
    ;Process
	if b_NewLogFile and FileExist(s_LogFile)
		FileDelete, %s_LogFile%
	ScriptLog(A_Now . "-" . ScriptName . "-(" . s_ScriptAction . "|" . b_DOS_EOLAction . "|" . b_NoBOMAction . "|" . s_FinalCode . ") has begun. " . A_UserName . "!------------------------------------------------------------------------------------------------------------------------------------------`n", 0)
	ScanFolder(s_WorkPath . "\*.*",s_WorkFileExt,b_Recursive)
	;Finish
	ScriptLog(A_Now . "-" . ScriptName . "-(" . s_ScriptAction . "|" . b_DOS_EOLAction . "|" . b_NoBOMAction . "|" . s_FinalCode . ") has finish. " . A_UserName . "!------------------------------------------------------------------------------------------------------------------------------------------`n", 0)
	MsgBox, 64, % ScriptName, % ScriptName "-(" . s_ScriptAction . "|" . b_DOS_EOLAction . "|" . b_NoBOMAction . "|" . s_FinalCode . ") has finish.`n" . A_UserName . "!" 
	Return
;==========================================================================================
;SUBRUTINES--------------------------------------------------------------------------------
ScriptLog(_s_Log, _i_TabDeep=0){ ;Log in screen (GUI 1:) and file the messages of script. Needs global s_LogFile and s_TextLog. No Return value.
	try{  ;Asegurate global s_LogFile access
		While (A_Index <= _i_TabDeep) ; Index with tab and | the message
			_s_Log:= % "`t|" . _s_Log
		s_TextLog:= % s_TextLog . _s_Log
	
		;With user validation screen output
			;MsgBox, 64, % A_ScriptName . "-" . A_ThisFunc . "-(" . A_LineNumber . ")" ,%s_TextLog%   ;Debug
		;Fast screen output (no scroll)
			;ToolTip, %_s_Log% ;, 5, 300
		;In GUI scroll display
			GuiControl,,s_TextLog, %s_TextLog% ;Search the control name with AutoHotKey Window Spy
			if !b_Debug {	; Not Debug mode. Too slow and keyboard use problems
				ControlSend, Edit3, ^{End}, %s_ScriptWinTitle% ;Assuming your edit control is Writtable and your window is %s_ScriptWinTitle%
				;ControlSend, s_TextLog, ^{End}, %s_ScriptWinTitle% ;Don't work
			}
		;Log to file
			FileAppend %_s_Log%, %s_LogFile%
	}
	catch e {
        MsgBox 16, % A_ThisFunc . "-(" . A_LineNumber . ")", % "Error catch " e.What ", which was called at line " e.Line ", in:`n|" s_LogFile "|"
        ExitApp            
    }		
	Return
}
ScanFolder(s_ListDir,_s_WorkFileExt,_b_Recursive){ ;Scan folder (recursive True/false) and process files with extensin in _s_WorkFileExt. No Return value.
	SplitPath, s_ListDir, s_FileName, s_Dir, s_Extension, s_NameNoExt, s_Drive  ;Not all variables are used. Only for documentation reason
	_s_Recursive:= ""
	if _b_Recursive
		_s_Recursive:= "R"
	;Set the Progress bar
	GuiControl, Show, i_ProgBarGlob
	Loop, Files, %s_ListDir%, F%_s_Recursive% ; Scan recursive the files in the present folders
    {
	    i_NumberOfFiles:= A_Index
		GuiControl,, s_ProgBarGlobTxt, %i_NumberOfFiles% 
		;MsgBox, % ComObjCreate("Scripting.FileSystemObject").GetFolder(s_Dir).Files.Count ; Number of files
		;MsgBox, % ComObjCreate("Shell.Application").NameSpace(s_Dir).Items.Count ;Number of Folders+Files
	}
	if b_VerboseLog
		ScriptLog(A_Now . "-" . A_ThisFunc . "-(" . s_ScriptAction . "|" . b_DOS_EOLAction . "|" . b_NoBOMAction . "|" . s_FinalCode . "): " . s_Dir . "|`n", 0)
			;MsgBox, 64, ScanFolder, ScanFolder info: |%s_ListDir%| found in:`n|%s_Dir%|   ;Debug
	Loop, Files, %s_Dir%\*, F%_s_Recursive% ;Scan present * files
	{    
	        GuiControl,, i_ProgBarGlob, % 100*A_Index/i_NumberOfFiles
			GuiControl,, s_ProgBarGlobTxt, %i_NumberOfFiles% 
			if A_LoopFileExt in %_s_WorkFileExt%
				ProcessFile(A_LoopFileFullPath) ; Process the present files with the implicated file extensions		
	}
	GuiControl, Hide, i_ProgBarGlob
	;GuiControl, Hide, s_ProgBarGlobTxt
Return
}
ProcessFile(s_FilePath){ ;Get the file in array and process it. No Return value.
	try{	
		WorkArr:=Array()
		WorkArrChanged:=False ; During the process File (if False=Not process, True=Process)
		SmartEncoding:= "Unknow"
			; Posible values:	"Unknow",
			;					"UTF-8_BOM", "UTF-8_No-BOM",
			;					"UTF-16_BE_BOM","UTF-16_BE_No-BOM","UTF-16_LE_BOM","UTF-16_LE_No-BOM",
			;					"UTF-32_BE_BOM","UTF-32_BE_No-BOM","UTF-32_LE_BOM","UTF-32_LE_No-BOM",
			;					"Win1252_By_SP_FR_GE_Diacritic", "Win1252_By_SP_Diacritic"
			;					"CP850_By_SP_FR_GE_Diacritic", "CP850_By_SP_Diacritic"
			;					"CP437_By_SP_FR_GE_Diacritic", "CP850_By_SP_Diacritic"
		if b_VerboseLog
			ScriptLog(A_Now . "-" . A_ThisFunc . "-(" . s_ScriptAction . "|" . b_DOS_EOLAction . "|" . b_NoBOMAction . "|" . s_FinalCode . "): " . s_FilePath . "|`n", 1)
		WorkArr:=BinArr_FromFile(s_FilePath)
		if b_DOS_EOLAction{ ;DOS_end_of_line assuration: Search and remplace (False=Not process, True=Process)
			; Before search and remplace comprobate presence of isolate CR or LF for DOS_EOL. Remenber DOS end of line in (CR+LF), UNIX(LF), Mac(CR).
											; CR(Carriage Return)=chr(13)=chr(0x0D);LF(Line Feed)=chr(10)=chr(0x0A)
			WorkArrDataPointer:=NumGet(ComObjValue(WorkArr) + 12 + (A_PtrSize==8 ? 4 : 0)) ; See ByteArray() doc
			WorkArrSize:=WorkArr.MaxIndex()+1 ;The index begin in 0
			;UNIX EOL: Isolate LF(Line Feed)=chr(10)=chr(0x0A)			
			NoDOS_EOLHintChar:=Chr(0x0A)
			NoDOS_EOLHintDetected:=False
			IndexOffset:=0
			While (IndexOffset < WorkArrSize){
				InBufResultIsolateEOL:=InBuf(WorkArrDataPointer, &NoDOS_EOLHintChar, WorkArrSize, 1, IndexOffset) ;Gets the more representative for the Log
				;ScriptLog("NoDOS_EOLHintDetected: " . NoDOS_EOLHintDetected . "| InBufResult LF: " . InBufResultIsolateEOL . "|" . WorkArr[InBufResultIsolateEOL-1] . "|`n",3) ;Debug
				if (InBufResultIsolateEOL = -1){ ;No more LF 
					break
				} else if (InBufResultIsolateEOL = 0){ ; Isolate LF at the begining.
					NoDOS_EOLHintDetected:=True
					break
				} else if (InBufResultIsolateEOL > 0) and (WorkArr[InBufResultIsolateEOL-1] <> 0x0D) { ;Isolate LF.
					NoDOS_EOLHintDetected:=True
					break
				}
				IndexOffset:=InBufResultIsolateEOL+1
			}
			if NoDOS_EOLHintDetected{ ;Isolate LF presents, UNIX_to_DOS_EOL conversion needed.
				if b_VerboseLog{
					ScriptLog(A_Now . "-" . A_ThisFunc . "-(" . s_ScriptAction . "|" . b_DOS_EOLAction . "|" . b_NoBOMAction . "|" . s_FinalCode . "): ", 1)
					ScriptLog("(| InBufResult LF: " . InBufResultIsolateEOL . "|): ", 0)
				}
				ScriptLog("UNIX EOL (Isolate LF) detected in: " . s_FilePath . "|`n", 0)
				WorkArr:=UNIX_to_DOS_EOL(WorkArr)
				WorkArrChanged:=True
			}
			;Mac EOL: Isolate CR(Carriage Return)=chr(13)=chr(0x0D)
			NoDOS_EOLHintChar:=Chr(0x0D)
			NoDOS_EOLHintDetected:=False			
			IndexOffset:=0
			While (IndexOffset < WorkArrSize){				
				InBufResultIsolateEOL:=InBuf(WorkArrDataPointer, &NoDOS_EOLHintChar, WorkArrSize, 1, IndexOffset) ;Gets the more representative for the Log
				;ScriptLog("NoDOS_EOLHintDetected: " . NoDOS_EOLHintDetected . "| InBufResult CR: " . InBufResultIsolateEOL . "|" . WorkArr[InBufResultIsolateEOL+1] . "|`n",3) ;Debug
				if (InBufResultIsolateEOL = -1){ ;No more CR 
					break
				} else if (InBufResultIsolateEOL = WorkArrSize){ ; Isolate CR at the end.
					NoDOS_EOLHintDetected:=True
					break
				} else if (InBufResultIsolateEOL > 0) and (WorkArr[InBufResultIsolateEOL+1] <> 0x0A) { ;Isolate CR
					NoDOS_EOLHintDetected:=True
					break
				}
				IndexOffset:=InBufResultIsolateEOL+1
			}			
			if NoDOS_EOLHintDetected { ;Isolate CR presents, MAC_to_DOS_EOL conversion needed.
				if b_VerboseLog{
					ScriptLog(A_Now . "-" . A_ThisFunc . "-(" . s_ScriptAction . "|" . b_DOS_EOLAction . "|" . b_NoBOMAction . "|" . s_FinalCode . "): ", 1)
					ScriptLog("(| InBufResult CR: " . InBufResultIsolateEOL . "|): ", 0)
				}
				ScriptLog("MAC EOL (Isolate CR) detected in: " . s_FilePath . "|`n", 0)
				WorkArr:=MAC_to_DOS_EOL(WorkArr)
				WorkArrChanged:=True
			}
		}
		if (b_NoBOMAction){ ;No_BOM action: Search and eliminate BOM UTF-8 (True=Not process, False=Process)
							;Encoding			Representacin (Hex)      Representacin (Char)
							;UTF-8					EF BB BF			 	﻿
			UTF8HintDetected := False ;Use for codification options
			if (WorkArr[0]=0xEF and WorkArr[1]=0xBB and WorkArr[2]=0xBF){ ; BOM UTF-8 present
				WorkArr:=BinArr_FromBinArray(WorkArr, 3) ;The third char is the good (index begin in 0)
				WorkArrChanged:=True
				if b_VerboseLog
						ScriptLog(A_Now . "-" . A_ThisFunc . "-(" . s_ScriptAction . "|" . b_DOS_EOLAction . "|" . b_NoBOMAction . "|" . s_FinalCode . "): ", 1)
				ScriptLog("UTF-8 BOM eliminated in: " . s_FilePath . "|`n", 0)
			}
		}
		if (s_FinalCode<>"Nothing"){ ;Posible values: Nothing|Win1552|CP850|NoBOM UTF-8|BOM UTF-8
			WorkArrDataPointer:=NumGet(ComObjValue(WorkArr) + 12 + (A_PtrSize==8 ? 4 : 0)) ; See ByteArray() doc
			WorkArrSize:=WorkArr.MaxIndex()+1   ;The index begin in 0			
			;Begin of Smart text codification detection
			;Detection by BOM, at least 4 chars.
			if (SmartEncoding=="Unknow") and (WorkArrSize>3){
				if (WorkArr[0]=0xEF and WorkArr[1]=0xBB and WorkArr[2]=0xBF){									;BOM UTF-8	EF BB BF 	﻿
					SmartEncoding:="UTF-8_BOM"				
				} else if (WorkArr[0]=0x00 and WorkArr[1]=0x00 and WorkArr[2]=0xFE and WorkArr[3]=0xFF){  		;BOM UTF-32 Big Endian      00 00 FE FF
					SmartEncoding:="UTF-32_BE_BOM"
				} else if (WorkArr[0]=0xFF and WorkArr[1]=0xFE and WorkArr[2]=0x00 and WorkArr[3]=0x00){  		;BOM UTF-32 Little Endian   FF FE 00 00 (always test before UTF-16 Little Endian) 
					SmartEncoding:="UTF-32_LE_BOM"
				} else if (WorkArr[0]=0xFE and WorkArr[1]=0xFF){ 												;BOM UTF-16 Big Endian	FE FF	
					SmartEncoding:="UTF-16_BE_BOM"
				} else if (WorkArr[0]=0xFF and WorkArr[1]=0xFE){ 												;BOM UTF-16 Little Endian   FF FE	
					SmartEncoding:="UTF-16_LE_BOM"
				}
			}						
			;Detection without BOM, at least 4 chars. Control Hints: UTF-16 and 32 contains Null chars. https://www.codeproject.com/Tips/672470/Simple-Character-Encoding-Detection
			if (SmartEncoding=="Unknow") and (WorkArrSize>3){
				if (WorkArr[0]=0x00 and WorkArr[1]=0x00){  				;No BOM UTF-32	Big Endian (00 00 XX XX)
					SmartEncoding:="UTF-32_BE_No-BOM"
				} else if (WorkArr[2]=0x00 and WorkArr[3]=0x00){  		;No BOM UTF-32 Little Endian (XX XX 00 00)
					SmartEncoding:="UTF-32_LE_No-BOM"
				} else if (WorkArr[0]=0x00){  			 				;No BOM UTF-16	Big Endian  (00 XX 00 XX)  (always test after UTF-32 Little Endian) 
						SmartEncoding:="UTF-16_BE_No-BOM"
				} else if (WorkArr[1]=0x00){					  		;No BOM UTF-16 Little Endian (XX 00 XX 00) (always test after UTF-32 Little Endian)
						SmartEncoding:="UTF-16_LE_No-BOM"
				}		
			}
			;Detection without BOM, Diacritic Hints: After dectect UTF-16 and 32 search the diferente diacritic codification in UTF-8 (two char) with Win1252,CP850 and CP437.
			if (SmartEncoding=="Unknow"){ ;Search the UTF-8 no BOM files with the first char hints for bypass the files without especial chars (for instance english) 
									 ;Con 2 Bytes=	De 0xC281   a 0xC2BF	( 37 char). C2= (Latin Capital Letter A With Circumflex): NO, es diacrtico francs y slo afecta a (¡) y (¿) en espaol Win1252.
				HintChar0:=Chr(0xC3) ;				De 0xC380   a 0xC3BF	( 64 char). Como comienzan por C3= (Latin Capital Letter A With Tilde) slo es diacrtico en portugus con Win1252.
				HintChar1:=Chr(0xC5) ;				De 0xC592   a 0xC5BE	(  7 char). Como comienzan por C5= (Latin Capital Letter A With Ring Above) slo es diacrtico en dans con Win1252.
                                     ;				De 0xC692 				(  1 char). Sera 0xC692=ƒ (Latin Capital Ligature Ae)+(Right Single Quotation Mark): NO, es diacrtico francs y no afecta espaol ...
				                     ;				De 0xCB86   a 0xCB9C	(  2 char). Son 0xCB86=ˆ / 0xCB9C=˜ (Latin Cap. E With Diaeresis)/+(Dagger)/+(Latin Small Ligature Oe): NO, es diacrtico francs ...
				                     ;Con 3 Bytes= 	De 0xE28093 a 0xE284A2	( 17 char). Como comienzan por E2= (Latin Small Letter A With Circumflex) seguidos 0x80= (Euro Sign), 0x82= (Single Low-9 Quot. Mark) 
									 ;													0x84= (Double Low-9 Quotation Mark): NO, es diacrtico francs y no hay ningn cdigo ... Win1252 con 3 digitos
				HintDetected:= (-1<>InBuf(WorkArrDataPointer,&HintChar0,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar1,WorkArrSize,1,0))
				if HintDetected ;UTF-8 code char presents.
					SmartEncoding:="UTF-8_No-BOM"
			}
			;Detection without BOM, Control Hints: The 5 ANSI_Win1252 control char don't discriminate between CP850 and CP437.
				; Chr(0x7F)	CP850= DEL	/	CP437= House	and ANSI_Win1252= Control character-1 (Delete)
				; Chr(0x81) CP850/CP437= 					and	ANSI_Win1252= Control character-1 (High Octet Preset)
				; Chr(0x8D)	CP850/CP437= 					and ANSI_Win1252= Control character-1 (Reverse Line Feed)
				; Chr(0x8F) CP850/CP437= 					and ANSI_Win1252= Control character-1 (Single-Shift Three)
				; Chr(0x90) CP850/CP437= 					and	ANSI_Win1252= Control character-1 (Device Control String)
                ; Chr(0x9D) CP850= 	/	CP437=  		and	ANSI_Win1252= Control character-1 (Operating System Command)
			;Detection without BOM, ANSI_Win1252 Diacritic Hints: "Win1252_By_SP_FR_GE_Diacritic",
			if (SmartEncoding=="Unknow"){ ;
				HintChar0:=Chr(0xAA) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar1:=Chr(0xBA) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar2:=Chr(0xBF) ; ANSI_Win1252=      /     CP805= +     /     CP437= +
				HintChar3:=Chr(0xC0) ; ANSI_Win1252=      /     CP805= +     /     CP437= +
				HintChar4:=Chr(0xC1) ; ANSI_Win1252=      /     CP805= -     /     CP437= -
				HintChar5:=Chr(0xC2) ; ANSI_Win1252=      /     CP805= -     /     CP437= -
				HintChar6:=Chr(0xC4) ; ANSI_Win1252=      /     CP805= -     /     CP437= -
				HintChar7:=Chr(0xC6) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar8:=Chr(0xC7) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar9:=Chr(0xC8) ; ANSI_Win1252=      /     CP805= +     /     CP437= +
				HintChar10:=Chr(0xC9) ; ANSI_Win1252=      /     CP805= +     /     CP437= +
				HintChar11:=Chr(0xCA) ; ANSI_Win1252=      /     CP805= -     /     CP437= -
				HintChar12:=Chr(0xCB) ; ANSI_Win1252=      /     CP805= -     /     CP437= -
				HintChar13:=Chr(0xCC) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar14:=Chr(0xCD) ; ANSI_Win1252=      /     CP805= -     /     CP437= -
				HintChar15:=Chr(0xCE) ; ANSI_Win1252=      /     CP805= +     /     CP437= +
				HintChar16:=Chr(0xCF) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar17:=Chr(0xD1) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar18:=Chr(0xD9) ; ANSI_Win1252=      /     CP805= +     /     CP437= +
				HintChar19:=Chr(0xDA) ; ANSI_Win1252=      /     CP805= +     /     CP437= +
				HintChar20:=Chr(0xDB) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar21:=Chr(0xDC) ; ANSI_Win1252=      /     CP805= _     /     CP437= _
				HintChar22:=Chr(0xDF) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar23:=Chr(0xE4) ; ANSI_Win1252=      /     CP805=      /     CP437= S
				HintChar24:=Chr(0xE6) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar25:=Chr(0xE7) ; ANSI_Win1252=      /     CP805=      /     CP437= t
				HintChar26:=Chr(0xE8) ; ANSI_Win1252=      /     CP805=      /     CP437= F
				HintChar27:=Chr(0xEC) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar28:=Chr(0xED) ; ANSI_Win1252=      /     CP805=      /     CP437= f
				HintChar29:=Chr(0xEE) ; ANSI_Win1252=      /     CP805=      /     CP437= e
				HintChar30:=Chr(0xEF) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar31:=Chr(0xF2) ; ANSI_Win1252=      /     CP805= =     /     CP437= =
				HintChar32:=Chr(0xF3) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar33:=Chr(0xF4) ; ANSI_Win1252=      /     CP805=      /     CP437= (
				HintChar34:=Chr(0xF9) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar35:=Chr(0xFA) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar36:=Chr(0xFB) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar37:=Chr(0xFC) ; ANSI_Win1252=      /     CP805=      /     CP437= 

				HintDetected:= (-1<>InBuf(WorkArrDataPointer,&HintChar0,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar1,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar2,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar3,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar4,WorkArrSize,1,0))				
								or (-1<>InBuf(WorkArrDataPointer,&HintChar5,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar6,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar7,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar8,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar9,WorkArrSize,1,0))
				If Not HintDetected
					HintDetected:= HintDetected
								or (-1<>InBuf(WorkArrDataPointer,&HintChar10,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar11,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar12,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar13,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar14,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar15,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar16,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar17,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar18,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar19,WorkArrSize,1,0))
				If Not HintDetected
					HintDetected:= HintDetected								
								or (-1<>InBuf(WorkArrDataPointer,&HintChar20,WorkArrSize,1,0))				
								or (-1<>InBuf(WorkArrDataPointer,&HintChar21,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar22,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar23,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar24,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar25,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar26,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar27,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar28,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar29,WorkArrSize,1,0))
				If Not HintDetected
					HintDetected:= HintDetected								
								or (-1<>InBuf(WorkArrDataPointer,&HintChar30,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar31,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar32,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar33,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar34,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar35,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar36,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar37,WorkArrSize,1,0))
				if HintDetected
					SmartEncoding:="Win1252_By_SP_FR_GE_Diacritic"
			}
			;Detection without BOM, ANSI_Win1252 Diacritic Hints: "Win1252_By_SP_Diacritic"
			if (SmartEncoding=="Unknow") and b_SP_Diacritic { ;			
				HintChar0:=Chr(0xD3) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar1:=Chr(0xE1) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar2:=Chr(0xF1) ; ANSI_Win1252=      /     CP805=      /     CP437= 
				HintChar3:=Chr(0x80) ; ANSI_Win1252=      /     CP805=      /     CP437= 

				HintDetected:= (-1<>InBuf(WorkArrDataPointer,&HintChar0,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar1,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar2,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar3,WorkArrSize,1,0))
				if HintDetected
					SmartEncoding:="Win1252_By_SP_Diacritic"
			}		
			;Detection without BOM, CP850 Diacritic Hints: "CP850_By_SP_FR_GE_Diacritic"
			if (SmartEncoding=="Unknow"){ ;			
				HintChar0:=Chr(0xB7) ; CP805=      /      ANSI_Win1252=      /     CP437= +
				HintChar1:=Chr(0xB5) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar2:=Chr(0xB6) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar3:=Chr(0xDE) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar4:=Chr(0xD8) ; CP805=      /      ANSI_Win1252=      /     CP437= +
				HintChar5:=Chr(0xE3) ; CP805=      /      ANSI_Win1252=      /     CP437= p
				
				HintDetected:= (-1<>InBuf(WorkArrDataPointer,&HintChar0,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar1,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar2,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar3,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar4,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar5,WorkArrSize,1,0))
				if HintDetected
					SmartEncoding:="CP850_By_SP_FR_GE_Diacritic"
			}
			;Detection without BOM, CP850 Diacritic Hints: "CP850_By_SP_Diacritic"
			if (SmartEncoding=="Unknow") and b_SP_Diacritic { ;			
				HintChar0:=Chr(0xD6) ; CP805=      /      ANSI_Win1252=      /     CP437= +
				HintChar1:=Chr(0xE0) ; CP805=      /      ANSI_Win1252=      /     CP437= a
				
				HintDetected:= (-1<>InBuf(WorkArrDataPointer,&HintChar0,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar1,WorkArrSize,1,0))
				if HintDetected
					SmartEncoding:="CP850_By_SP_Diacritic"
			}
			;Detection without BOM, CP850 and CP437 common Diacritic Hints: "CP850_By_SP_FR_GE_Diacritic" or "CP437_By_SP_FR_GE_Diacritic"
			if (SmartEncoding=="Unknow"){ ;			
				HintChar0:=Chr(0xAD) ; CP805=      /      ANSI_Win1252= SHY     /     CP437= 
				HintChar1:=Chr(0xA6) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar2:=Chr(0xA7) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar3:=Chr(0xA8) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar4:=Chr(0x8E) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar5:=Chr(0x92) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar6:=Chr(0x90) ; CP805=      /      ANSI_Win1252= DCS     /     CP437= 
				HintChar7:=Chr(0xA5) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar8:=Chr(0x99) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar9:=Chr(0x9A) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar10:=Chr(0x85) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar11:=Chr(0xA0) ; CP805=      /      ANSI_Win1252= NBSP     /     CP437= 
				HintChar12:=Chr(0x83) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar13:=Chr(0x84) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar14:=Chr(0x91) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar15:=Chr(0x87) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar16:=Chr(0x8A) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar17:=Chr(0x82) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar18:=Chr(0x88) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar19:=Chr(0x89) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar20:=Chr(0x8D) ; CP805=      /      ANSI_Win1252= RI     /     CP437= 
				HintChar21:=Chr(0x8B) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar22:=Chr(0xA4) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar23:=Chr(0x95) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar24:=Chr(0xA2) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar25:=Chr(0x93) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar26:=Chr(0x94) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar27:=Chr(0x97) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar28:=Chr(0x96) ; CP805=      /      ANSI_Win1252=      /     CP437= 
				HintChar29:=Chr(0x81) ; CP805=      /      ANSI_Win1252= HOP     /     CP437= 

				HintDetected:= (-1<>InBuf(WorkArrDataPointer,&HintChar0,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar1,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar2,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar3,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar4,WorkArrSize,1,0))				
								or (-1<>InBuf(WorkArrDataPointer,&HintChar5,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar6,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar7,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar8,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar9,WorkArrSize,1,0))
				If Not HintDetected
					HintDetected:= HintDetected
								or (-1<>InBuf(WorkArrDataPointer,&HintChar10,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar11,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar12,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar13,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar14,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar15,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar16,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar17,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar18,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar19,WorkArrSize,1,0))
				If Not HintDetected
					HintDetected:= HintDetected								
								or (-1<>InBuf(WorkArrDataPointer,&HintChar20,WorkArrSize,1,0))				
								or (-1<>InBuf(WorkArrDataPointer,&HintChar21,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar22,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar23,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar24,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar25,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar26,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar27,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar28,WorkArrSize,1,0))
								or (-1<>InBuf(WorkArrDataPointer,&HintChar29,WorkArrSize,1,0))

				if HintDetected {
					if b_ChooseCP850
						SmartEncoding:="CP850_By_SP_FR_GE_Diacritic"
					if b_ChooseCP437
						SmartEncoding:="CP437_By_SP_FR_GE_Diacritic"
				}
			}
			;Detection without BOM, CP850 and CP437 common Diacritic Hints: "CP850_By_SP_Diacritic" or "CP437_By_SP_Diacritic"
			if (SmartEncoding=="Unknow") and b_SP_Diacritic { ;
				HintChar0:=Chr(0x9C) ; CP805=      /      ANSI_Win1252=      /     CP437= 				
				HintDetected:= (-1<>InBuf(WorkArrDataPointer,&HintChar0,WorkArrSize,1,0))
				if HintDetected {
					if b_ChooseCP850
						SmartEncoding:="CP850_By_SP_Diacritic"
					if b_ChooseCP437
						SmartEncoding:="CP437_By_SP_Diacritic"
				}
			}
			;Smart detection result
			if b_VerboseLog
				ScriptLog(A_Now . "-" . A_ThisFunc . "-(" . s_ScriptAction . "|" . b_DOS_EOLAction . "|" . b_NoBOMAction . "|" . s_FinalCode . "): ", 1)
			ScriptLog(SmartEncoding . " detected in: " . s_FilePath . "|`n", 0)

			; Posible values SmartEncoding:	"Unknow",
			;					"UTF-8_BOM", "UTF-8_No-BOM",
			;					"UTF-16_BE_BOM","UTF-16_BE_No-BOM","UTF-16_LE_BOM","UTF-16_LE_No-BOM",
			;					"UTF-32_BE_BOM","UTF-32_BE_No-BOM","UTF-32_LE_BOM","UTF-32_LE_No-BOM",
			;					"Win1252_By_SP_FR_GE_Diacritic", "Win1252_By_SP_Diacritic"
			;					"CP850_By_SP_FR_GE_Diacritic", "CP850_By_SP_Diacritic"
			;					"CP437_By_SP_FR_GE_Diacritic", "CP850_By_SP_Diacritic"
			; Posible values for Final Code (See Rosetta) https://docs.microsoft.com/en-us/sql/ado/reference/ado-api/charset-property-ado
			; The ADODB.Stream component of the Charset property value set: http://www.scriptcode.net/thread-49078-1-1.html
			if ((SmartEncoding<>"Unknow") or b_ForceCodeUnknow) and (s_FinalCode<>"Nothing"){ ; and (s_ScriptAction == "Process") {
				InCharset:="_autodetect_all"
				if (SmartEncoding=="UTF-8_BOM") or (SmartEncoding=="UTF-8_No-BOM")
					InCharset:="UTF-8"
				if (SmartEncoding=="UTF-16_BE_BOM") or (SmartEncoding=="UTF-16_BE_No-BOM")
					InCharset:="unicodeFEFF"	; "unicode" and "UTF-16" only works with BOM (bad chars in No_BOM).
												; "UTF-16BE" only works with No_BOM (show as bad chars the BOM).  Not recogniced: "UTF-16_BE", "BigEndianUnicode", "u16BE", "UTF-16 (BE)"
				if (SmartEncoding=="UTF-16_LE_BOM") or (SmartEncoding=="UTF-16_LE_No-BOM")
					InCharset:="unicodeFFFE"	; Idem "UTF-16LE"
				;Not valid ADO Charset found for UTF-32. Avoid conversion.
				;if (SmartEncoding=="UTF-32_BE_BOM") or (SmartEncoding=="UTF-32_BE_No-BOM")  	; Mal
				;	InCharset:=				; NOK:"unicode"(empty),"_autodetect_all"(empty), "unicodeFEFF"(empty)
				;									; Not recogniced: "UTF-32_BE", "utf-32", "UTF-32", "UTF32", "UTF-32BE", "UTF_32BE", "UTF32BE", "UTF-32 (BE)", "unicodeFFFE0000", "_autodetect"
				;if (SmartEncoding=="UTF-32_LE_BOM") or  (SmartEncoding=="UTF-32_LE_No-BOM")  	; Mal
				;	InCharset:=				;"UTF-32_LE"
				if (SmartEncoding=="Win1252_By_SP_FR_GE_Diacritic") or  (SmartEncoding=="Win1252_By_SP_Diacritic")
					InCharset:="Windows-1252"		; or "iso-8859-1"
				if (SmartEncoding=="CP850_By_SP_FR_GE_Diacritic") or  (SmartEncoding=="CP850_By_SP_Diacritic")
					InCharset:="IBM850"		; Not recogniced: "CP850", "850"
				if (SmartEncoding=="CP437_By_SP_FR_GE_Diacritic") or  (SmartEncoding=="CP437_By_SP_Diacritic")
					InCharset:="IBM437"		; Not recogniced: "CP437" For reading the string is the same. is the file reader the responsable of show "letter" or "box" in no common chars.

				;Not valid ADO Charset found for UTF-32. Avoid conversion.
				if Not ((SmartEncoding=="UTF-32_BE_BOM") or (SmartEncoding=="UTF-32_BE_No-BOM") or (SmartEncoding=="UTF-32_LE_BOM") or  (SmartEncoding=="UTF-32_LE_No-BOM")) { 
					TmpStr:=BinArr_ToString(WorkArr,InCharset)
					if b_ShowDecodeText{ ; Show the decode text
						MsgBox 0x21, Confirm Good Detection? [INTRO=Yes/ESC=No], % s_FilePath . "`n[" SmartEncoding "/" InCharset "]`n|" TmpStr "|"
						IfMsgBox Cancel ;Smart detention have fail
							SmartEncoding:="Fail"					
						if b_VerboseLog
							ScriptLog(A_Now . "-" . A_ThisFunc . "-(" . s_ScriptAction . "|" . b_DOS_EOLAction . "|" . b_NoBOMAction . "|" . s_FinalCode . "): ", 1)
						ScriptLog(SmartEncoding . " smart encoding set by user in: " . s_FilePath . "|`n", 0)
					}
					if (s_FinalCode=="UTF-8_BOM") and (SmartEncoding<>"UTF-8_BOM") and (SmartEncoding<>"Fail"){
						WorkArr:=BinArr_FromString(TmpStr,"UTF-8", True)
						WorkArrChanged:=True						
					} else if (s_FinalCode=="UTF-8_No-BOM") and (SmartEncoding<>"UTF-8_No-BOM") and (SmartEncoding<>"Fail"){
						WorkArr:=BinArr_FromString(TmpStr,"UTF-8", False)
						WorkArrChanged:=True						
					} else if (s_FinalCode=="UTF-16BE_BOM") and (SmartEncoding<>"UTF-16BE_BOM") and (SmartEncoding<>"Fail"){
						WorkArr:=BinArr_FromString(TmpStr,"unicodeFEFF", True)
						WorkArrChanged:=True						
					} else if (s_FinalCode=="UTF-16BE_No-BOM") and (SmartEncoding<>"UTF-16BE_No-BOM") and (SmartEncoding<>"Fail"){
						WorkArr:=BinArr_FromString(TmpStr,"unicodeFEFF", False)
						WorkArrChanged:=True						
					} else if (s_FinalCode=="UTF-16LE_BOM") and (SmartEncoding<>"UTF-16LE_BOM") and (SmartEncoding<>"Fail"){
						WorkArr:=BinArr_FromString(TmpStr,"unicodeFFFE", True)
						WorkArrChanged:=True						
					} else if (s_FinalCode=="UTF-16LE_No-BOM") and (SmartEncoding<>"UTF-16LE_No-BOM") and (SmartEncoding<>"Fail"){
						WorkArr:=BinArr_FromString(TmpStr,"unicodeFFFE", False)
						WorkArrChanged:=True
					} else if (s_FinalCode=="Win1552") and (InCharset<>"Windows-1252") and (SmartEncoding<>"Fail"){
						WorkArr:=BinArr_FromString(TmpStr,"Windows-1252", False)
						WorkArrChanged:=True
					} else if (s_FinalCode=="CP850") and ((InCharset<>"IBM850") or (InCharset<>"IBM437")) and (SmartEncoding<>"Fail"){	; Reading CP8505 or CP437 is the same for the string conversion.
						WorkArr:=BinArr_FromString(TmpStr,"IBM850", False)																; Is the file reader the responsable of show "letter" or "box".
						WorkArrChanged:=True
					}
				}
			}
		}
		;MsgBox 64, % A_ThisFunc . "-(" . A_LineNumber . ")", % WorkArr.MaxIndex() . "`n|" s_FilePath "|"
		if (s_ScriptAction == "Process" and WorkArrChanged == True){ 	; Script action: =Analyze (not files changes), =Process overwritting files
			if FileExist(s_FilePath)
				FileDelete, %s_FilePath%			
			BinArr_ToFile(WorkArr, s_FilePath) 
		}
	}
	catch e {
        MsgBox 16, % A_ThisFunc . "-(" . A_LineNumber . ")", % "Error catch " e.What ", which was called at line " e.Line ", in:`n|" s_FilePath "|"
        ExitApp            
    }
	Return
}
UNIX_to_DOS_EOL(_InArray){ ;Returns array with DOS end of line (CR+LF) from array with UNIX EOL: Isolate LF(Line Feed)=chr(10)=chr(0x0A)
	; With DOS end of line in (CR+LF), UNIX(LF), Mac(CR).
	; CR(Carriage Return)=chr(13)=chr(0x0D);LF(Line Feed)=chr(10)=chr(0x0A)
	;Definition Rosetta Stone for UNIX(LF) (IMPORTANT if the file is DOS we get CR+CR+LF)
	RosettaArray:= []  ;[TokenBad s_ing uChar to search, TokenGood s_ing uChar to replace, int action 1=replace, 2=only alert ]
	RosettaArray[0]:=[Chr(0x0A), Chr(0x0D) . Chr(0x0A), 1]	;LF->CR+LF
	TmpArray:=RemplaceToken(_InArray,RosettaArray,"DOS_EOL")	
	;Definition Rosetta Stone for repare the DOS file (CR+CR+LF)
	RosettaArray[0]:=[Chr(0x0D) . Chr(0x0D) . Chr(0x0A), Chr(0x0D) . Chr(0x0A), 1] ;CR+CR+LF->CR+LF               
	TmpArray:=RemplaceToken(TmpArray,RosettaArray,"DOS_EOL")	
	Return TmpArray
}
MAC_to_DOS_EOL(_InArray){ ;Returns array with DOS end of line (CR+LF) from array with Mac EOL: Isolate CR(Carriage Return)=chr(13)=chr(0x0D)	
	; With DOS end of line in (CR+LF), UNIX(LF), Mac(CR).
	; CR(Carriage Return)=chr(13)=chr(0x0D);LF(Line Feed)=chr(10)=chr(0x0A)
	;Definition Rosetta Stone for Mac(CR) (IMPORTANT if the file is DOS we get CR+LF+LF)
	RosettaArray:= []  ;[TokenBad s_ing uChar to search, TokenGood s_ing uChar to replace, int action 1=replace, 2=only alert ]
	RosettaArray[0]:=[Chr(0x0D), Chr(0x0D) . Chr(0x0A), 1] ;CR->CR+LF               
	TmpArray:=RemplaceToken(_InArray,RosettaArray,"DOS_EOL")	
	;Definition Rosetta Stone for for repare the DOS file (CR+LF+LF)
	RosettaArray[0]:=[Chr(0x0D) . Chr(0x0A) . Chr(0x0A), Chr(0x0D) . Chr(0x0A), 1] ;CR->CR+LF               
	TmpArray:=RemplaceToken(TmpArray, RosettaArray, "DOS_EOL")	
	Return TmpArray
}
RemplaceToken(InArray, RosettaArray, TokenType){ ;Returns InArray with the RosettaArray Token processed, normaly rempalced. If TokenType = "NO_BOM"=Only the first place of InArray are analysed. "DOS_EOL"=No Log activity. "UTF8_WIN1252"=Nothing especial, full analyse.
	WorkArray:= ComObjArray(0x11, 2*(InArray.MaxIndex()+1)) ; VT_ARRAY(0x2000)+VT_UI1(8-bit unsigned int 0x11) = BinArr(0x2011=8209), Dimension Size=0x100000=1024x1024=1048575
	;MsgBox 64, % A_ThisFunc . "-(" . A_LineNumber . ")", % "InArray is object:" . IsObject(InArray) . "`n" . "OutArray is object:" . IsObject(OutArray) . "`n"
	;MsgBox 64, % A_ThisFunc . "-(" . A_LineNumber . ")", % "InArray ComObjType:" . ComObjType(InArray) . "`n" . "OutArray ComObjType:" . ComObjType(OutArray) . "`n"
	;MsgBox 64, % A_ThisFunc . "-(" . A_LineNumber . ")", % "InArray MaxIndex=" . InArray.MaxIndex() . "`n" . "OutArray MaxIndex=" . OutArray.MaxIndex() . "`n"

	;Loop to read the input characters
	indexCharOut:=0
	indexCharIn:=0
	GuiControl, Show, i_ProgBarLocal
	GuiControl, Show, s_ProgBarLocalTxt
	i_ArrayMaxIndex:= InArray.MaxIndex()
	While (indexCharIn <= i_ArrayMaxIndex){
	    GuiControl,, i_ProgBarLocal, % 100*indexCharIn/i_ArrayMaxIndex
	    GuiControl,, s_ProgBarLocalTxt, % Format("{1:0.1f}", 100*indexCharIn/i_ArrayMaxIndex) "`%"
		InArrayChar:=InArray[indexCharIn]
		if b_Debug and b_VerboseLog ; Debug
			ScriptLog(A_ThisFunc . "-(" . A_LineNumber . ")-" . TokenType . " Input char #" . indexCharIn . " | is |" . InArrayChar . "|" . chr(InArrayChar) . "|`n", 4)
		;Loop to compare with the possible TokenBad s_ing values
		TokenDetected:=False ;For asure the copy of the char if not token analyse, for instance for "NO_BOM".
		indexRosetta :=0
		i_RosettaMaxIndex:= RosettaArray.MaxIndex()
		While (indexRosetta <= i_RosettaMaxIndex) and !(TokenType == "NO_BOM" and indexCharIn > 0){ ;For "NO_BOM" the token analyse it's necesary for the first place (index=0) but it must copy the rest of chars.
			Token:=RosettaArray[indexRosetta,1]
			TokenOut:=RosettaArray[indexRosetta,2]
			TokenAction:=RosettaArray[indexRosetta,3]
			if b_Debug and b_VerboseLog ; Debug
				ScriptLog(A_ThisFunc . "-(" . A_LineNumber . ")-" . TokenType . " Token #" . indexRosetta . "| is (In/Out/Action) |" . Token . "|" . TokenOut . "|" . TokenAction . "|`n", 5)
			;Loop to compare all the characters in the Token
			TokenDetected:=True
			indexTokenChar:=0
			While (indexTokenChar < StrLen(Token)){
				indexFollowingCharIn:=indexCharIn+indexTokenChar ;The index to the next needed char
				if (indexFollowingCharIn > InArray.MaxIndex()){ ;The next needed char is out of range
					TokenDetected:=False
					break ;Go to the next Token	
				}
				InArrayChar:=InArray[indexFollowingCharIn]
				TokenChar:=Asc(SubStr(Token,indexTokenChar+1,1))
				if b_Debug and b_VerboseLog ; Debug
					ScriptLog(A_ThisFunc . "-(" . A_LineNumber . ")-" . TokenType . " In_Char #" . indexCharIn+indexTokenChar . " |is|" . InArrayChar . "|" . chr(InArrayChar) . "|<->|" . TokenChar . "|`n", 5)
				if (InArrayChar <> TokenChar){ ;Not "char" match detected
					TokenDetected:=False
					break ;Go to the next Token
				}
				indexTokenChar+=1 ;Go to the next "char" of the actual Token
			}
			if (TokenDetected=True){ ;Entire Token detected matching, not necesary more Token
				if (TokenAction=1){ ;Action 1=replace, 2=only alert
					indexTokenChar:=0
					if (0 = StrLen(TokenOut)) and (b_VerboseLog and (TokenType <> "DOS_EOL")) ;In "DOS_EOL"=No normal Log activity.
							ScriptLog(TokenType . "-Token Match:" . Token . "| remplaced by:" . TokenOut . "|`n", 2)
					While (0 < StrLen(TokenOut) and indexTokenChar < StrLen(TokenOut)){
						TokenChar:=Asc(SubStr(TokenOut,indexTokenChar+1,1))
						if (b_VerboseLog and (TokenType <> "DOS_EOL")) ;In "DOS_EOL"=No normal Log activity.
							ScriptLog(TokenType . "-Token Match:" . Token . "| remplaced by:" . TokenOut . "|`n", 2)
						WorkArray[indexCharOut]:=TokenChar
						indexTokenChar+=1
						indexCharOut+=1
					}
				} else if (TokenAction=2){ ;Action 2=only alert
					if (b_VerboseLog and (TokenType <> "DOS_EOL")) ;In "DOS_EOL"=No normal Log activity.
						ScriptLog(TokenType . "-ALERT-Token Match:"  . Token . "| ALERT.`n", 2)										
				}	
				break ;Because the Token is detected matching it's not necesary comprobate moreToken, go to the next input characters	
			}	
			indexRosetta+=1 ;Go to the next possible "active" s_ing values (BOM code in thes case)
			;MsgBox 64, % A_ThisFunc . "-(" . A_LineNumber . ")", % "In_char #" . indexCharIn . " |is|" . InArray[indexCharIn] . "|<->|" . TokenChar . "|"
		}
		if (TokenDetected=True){ ;Entire Token detected matching
			indexCharIn+=StrLen(Token) ;Skip the Token length
		} else { ;Not Token match detected
			if b_Debug and b_VerboseLog ; Debug
				ScriptLog(A_ThisFunc . "-(" . A_LineNumber . ")-" . TokenType . " Not Token match. IndexCharOut #" . indexCharOut . " | is |" . InArray[indexCharIn] . "|" . chr(InArray[indexCharIn]) . "|`n", 3)	
			WorkArray[indexCharOut]:=InArray[indexCharIn]
			indexCharOut+=1
			indexCharIn+=1
		}
	}
	;Set OutArray to the result with the proper length (SetCapacity don't work, may be with DllCall("SafeArrayRedim" ...
	OutArray:= ComObjArray(0x11, indexCharOut) ; VT_ARRAY(0x2000)+VT_UI1(8-bit unsigned int 0x11) = BinArr(0x2011=8209)
	Loop % OutArray.MaxIndex()+1
			OutArray[A_Index-1]:=WorkArray[A_Index-1]
	;MsgBox 64, % A_ThisFunc . "-(" . A_LineNumber . ")", % "In  length:" InArray.MaxIndex() . "`n"
	;MsgBox 64, % A_ThisFunc . "-(" . A_LineNumber . ")", % "Out length:" OutArray.MaxIndex() . "`n"
	;MsgBox 64, % A_ThisFunc . "-(" . A_LineNumber . ")", % BinArr_Tos_ing(OutArray) . "`n"
	GuiControl, Hide, i_ProgBarLocal
	GuiControl, Hide, s_ProgBarLocalTxt
	Return OutArray
}
;==========================================================================================
;AutoHotKey Snipets------------------------------------------------------------------------ 
BinArr_FromString(str, Encoding := "UTF-8", BOM := False){ ;Create binary array from file/str using ADODB.Stream https://autohotkey.com/boards/viewtopic.php?f=6&t=7645 https://gist.github.com/tmplinshi/a97d9a99b9aa5a65fd20
											 ;Possible Encoding (See registry HKEY_CLASSES_ROOT\MIME\Database\Charset ):_autodetect_all, "CP850", "UTF-8"=UTF-8 with BOM, "Windows-1252"="ANSI", "Unicode"=UTF16LE,"iso-8859-1"=
											;https://docs.microsoft.com/en-us/sql/ado/reference/ado-api/charset-property-ado 
	oADO := ComObjCreate("ADODB.Stream")
	oADO.Type := 2 ; adTypeText
	oADO.Mode := 3 ; adModeReadWrite
	oADO.Open
	oADO.Charset := Encoding
	oADO.WriteText(str)
	oADO.Position := 0
	oADO.Type := 1 ; adTypeBinary
	if (Not BOM and (Encoding == "UTF-8")) ;Skip UTF-8 BOM	(EF BB BF)
		oADO.Position := 3
	if (Not BOM and ((Encoding == "unicodeFEFF") or (Encoding == "unicodeFFFE"))) ;Skip UTF-16 BOM UTF-16 Big Endian (FE FF) or UTF-16 Little Endian (FF FE)
		oADO.Position := 2
	return oADO.Read, oADO.Close
}
BinArr_FromFile(FileName){
	oADO := ComObjCreate("ADODB.Stream")
	oADO.Type := 1 ; adTypeBinary
	oADO.Open
	oADO.LoadFromFile(FileName)
	return oADO.Read, oADO.Close
}
BinArr_FromBinArray(_BinArray, BeginPosition:=0){
	oADO := ComObjCreate("ADODB.Stream")
	oADO.Type := 1 ; adTypeBinary
	oADO.Mode := 3 ; adModeReadWrite
	oADO.Open
	oADO.Write(_BinArray)
	oADO.Position := BeginPosition
	return oADO.Read, oADO.Close
}
BinArr_Join(Arrays*){
	oADO := ComObjCreate("ADODB.Stream")
	oADO.Type := 1 ; adTypeBinary
	oADO.Mode := 3 ; adModeReadWrite
	oADO.Open
	For i, arr in Arrays
		oADO.Write(arr)
	oADO.Position := 0
	return oADO.Read, oADO.Close
}
BinArr_ToString(BinArr, Encoding := "UTF-8"){
	oADO := ComObjCreate("ADODB.Stream")
	oADO.Type := 1 ; adTypeBinary
	oADO.Mode := 3 ; adModeReadWrite
	oADO.Open
	oADO.Write(BinArr)
	oADO.Position := 0
	oADO.Type := 2 ; adTypeText
	oADO.Charset  := Encoding 
	return oADO.ReadText, oADO.Close
}
BinArr_ToFile(BinArr, FileName){
	oADO := ComObjCreate("ADODB.Stream")
	oADO.Type := 1 ; adTypeBinary
	oADO.Open
	oADO.Write(BinArr)
	oADO.SaveToFile(FileName, 2)
	oADO.Close
}
InBuf(haystackAddr, needleAddr, haystackSize, needleSize, StartOffset=0){ ;Machine code binary buffer searching regardless of NUL. Returns either index position (begin 0) of 'needle' inside 'haystack' or -1 if not found. https://autohotkey.com/board/topic/23627-machine-code-binary-buffer-searching-regardless-of-null/
 	Static fun:=""
	IfEqual,fun,
	{
      h=
      ( LTrim join
         5589E583EC0C53515256579C8B5D1483FB000F8EC20000008B4D108B451829C129D9410F8E
         B10000008B7D0801C78B750C31C0FCAC4B742A4B742D4B74364B74144B753F93AD93F2AE0F
         858B000000391F75F4EB754EADF2AE757F3947FF75F7EB68F2AE7574EB628A26F2AE756C38
         2775F8EB569366AD93F2AE755E66391F75F7EB474E43AD8975FC89DAC1EB02895DF483E203
         8955F887DF87D187FB87CAF2AE75373947FF75F789FB89CA83C7038B75FC8B4DF485C97404
         F3A775DE8B4DF885C97404F3A675D389DF4F89F82B45089D5F5E5A595BC9C2140031C0F7D0EBF0
      )
      VarSetCapacity(fun,StrLen(h)//2)
      Loop % StrLen(h)//2
         NumPut("0x" . SubStr(h,2*A_Index-1,2), fun, A_Index-1, "Char")
   }
   Return DllCall(&fun
      , "uint",haystackAddr, "uint",needleAddr
      , "uint",haystackSize, "uint",needleSize
      , "uint",StartOffset)
}
;==========================================================================================
;General Documentation:
;--------------------------------------------------------------------------------
;--------------------------------------------------------------------------------
;==========================================================================================
;Program Documentation:
;COM Object Reference [AutoHotkey v1.1+]. Purpose: to create a basic reference for commonly used COM objects 
;		https://autohotkey.com/board/topic/56987-com-object-reference-autohotkey-v11/
;ComObjType() : Retrieves type information from a COM object.
;		https://autohotkey.com/docs/commands/ComObjType.htm
;ComObjArray() : Creates a SafeArray for use with COM.
;       https://autohotkey.com/docs/commands/ComObjArray.htm
;ByteArray() https://autohotkey.com/boards/viewtopic.php?t=6967       
;		SAFEARRAY: Estructura especial, heredada directamente de COM, donde se guarda informacin sobre la matriz, como dimensiones, tamao de elementos o direccin de memoria donde se almacenan sus elementos.)
;		https://msdn.microsoft.com/en-us/library/ms221482.aspx
;Stream Object (ADO): Represents a stream of binary data or text.
;		https://msdn.microsoft.com/en-us/library/ms675032(v=vs.85).aspx
;TF: Textfile & String Library for AutoHotkey [lib] - v3.6: A "Swiss Army Knife" library for Text (files)
;       https://github.com/hi5/TF
;ComObjType: https://autohotkey.com/docs/commands/ComObjType.htm
;The VarType function returns a value that indicates the subtype of a specified variable.
;The VarType function can return one of the following values:
;0 = vbEmpty - Indicates Empty (uninitialized)
;1 = vbNull - Indicates Null (no valid data)
;2 = vbInteger - Indicates an integer
;3 = vbLong - Indicates a long integer
;4 = vbSingle - Indicates a single-precision floating-point number
;5 = vbDouble - Indicates a double-precision floating-point number
;6 = vbCurrency - Indicates a currency
;7 = vbDate - Indicates a date
;8 = vbString - Indicates a string
;9 = vbObject - Indicates an automation object
;10 = vbError - Indicates an error
;11 = vbBoolean - Indicates a boolean
;12 = vbVariant - Indicates a variant (used only with arrays of Variants)
;13 = vbDataObject - Indicates a data-access object
;17 = vbByte - Indicates a byte
;8192 = 0x2000 = vbArray - Indicates an array
;Note: If the variable is an array VarType() returns 8192 + VarType(array_element). Example: for an array of integer VarType() will return 8192 + 2 = 8194.

